library CodeGen;

uses
  Windows,
  kol,
  CGTShare in '..\CGTShare.pas',
  Errors in 'Errors.pas';

{$define debug}
{$define called_func}      //    -
{$define called_func_sec}  //    -,     
{$define processed_func}   //     -
{$define print_lexem}      //   print
{$define call_func}
{$define event_call}       //    
{$define processed_unit}   //    unit-
{$define init_section}     //   ,  init     
{$define _blocks_}         //    

(*
**   todo
**     -    data_array
**
**
*)

type
  TCGrec = record
    Code:PStrList;
  end;
  PCGrec = ^TCGrec;

  TData = record
    data_type:byte;
    idata:integer;
    sdata:string;
    rdata:real;
  end;
  PScData = ^TScData;
  PScArray = ^TScArray;
  TScData = object
    private
     value:TData;
    public
     procedure SetValue(const val:string); overload;
     procedure SetValue(val:integer); overload;
     procedure SetValue(val:real); overload;
     procedure SetValue(val:boolean); overload;
     procedure SetValue(const val:TData); overload;
     procedure SetValue(const val:string; tp:byte); overload;
     procedure AddValue(const val:TScData);
     function GetType:byte;
     procedure SetType(tp:byte);
     procedure GetValue(var val:Tdata);
     function toStr:string;
     function toInt:integer;
     function toReal:real;
     function toCode:string;
     function toArray:PScArray;

     function isEmpty:boolean;
     //function isStr:boolean;
     //function isInt:boolean;
     //function isReal:boolean;
  end;
  TScArray = record
    Items:array of TScData;
    Count:integer;
  end;

  TBlockStack = class
    private
      Items:array of record
         LType:word;
         skip:boolean;
         userdata:pointer;
      end;
      Count:word;
    public
      procedure Push(_skip:boolean; _Type:word; _userdata:pointer);
      function Pop(var _Skip:boolean;var _Type:word; var _userdata:pointer):boolean;
      procedure Clear;
  end;
  TCodeBlock = class
   private
    FItems:PStrListEx;
    Current:PStrList;
    FCurName:string;
    FGen:integer;

    function Find(const Name:string):PStrList;
    function GetItems(Index:integer):PStrList;
    function GetCount:integer;
   public
    constructor Create;
    destructor Destroy; override;
    procedure Reg(const Name:string);
    function RegGen:string;
    function Select(const Name:string):string;
    procedure Delete(const Name:string);
    procedure CopyTo(const Source, Dest:string);
    procedure Merge(List:PStrList);
    procedure Add(const text:string);
    procedure Paste(const text:string);
    procedure AddStrings(list:PStrList);
    function Empty:boolean;
    function AsText:string;

    procedure SaveToFile(const fname, name:string);

    property Items[Index:integer]:PStrList read GetItems;
    property Count:integer read GetCount;
    property CurBlockName:string read FCurName;
  end;
  TArgs = class;
  TFuncData = record
    Line:integer;
    Skeep,fskeep:boolean;
    inFunc:boolean;
    fsection:integer;
    _args_:TArgs;
    _data_saved_:TScData;
  end;
  PFuncData = ^TFuncData;

  TFuncState = record
    Line:string;
    LineIndex,LPos:integer;
    Token,RToken:string;
    TokType:byte;
    fData:PFuncData;
  end;
  PFuncState = ^TFuncState;
  TArgs = class
   private
    FItems:PStrListEx;

    function GetValues(index:integer):PScData;
    function GetNames(index:integer):string;
    procedure SetNames(index:integer; const Value:string);
    function GetCount:integer;
   public
    constructor Create;
    destructor Destroy; override;
    function AddArg(const Name:string):PScData;

    function find(const name:string):PScData;

    property Names[Index:integer]:string read GetNames write SetNames;
    property Values[Index:integer]:PScData read GetValues;
    property Count:integer read GetCount;
  end;
  TParser = class
    private
     VarList:TArgs;          //   
     funcList:PStrListEx;
     procedure Debug(const text:string; color:TColor = clBlack);
     function isDebug:boolean;
     function ToSection(var sec:integer):boolean;
     procedure SaveState(state:PFuncState);
     procedure LoadState(state:PFuncState);
     function _points_count_(ptype:byte):integer;
    protected
     fData:PFuncData;
     _data_return_:TScData;

     Token,RToken:string;
     TokType:byte;
     uname:string;
     Line:string;
     Lines:PStrList;
     LineIndex,LPos:integer;
     oldLineIndex,oldLPos:integer;
     cgt:PCodeGenTools;
     el:id_element;
     BlockStack:TBlockStack;
     codeb:TCodeBlock;
     level:integer;
     _err_:boolean;
     section:integer;    //    

     function ReadLine:boolean;
     procedure AddError(const Error:string);
     procedure Start(const fname:string; Args:TArgs);
     function ReadPoint(lp:id_point):TScData;

     procedure SumVar(var _var_:TScData; const var2:TScData);

     procedure ReadValue(const Name:string; var _var_:TScData; null_data:boolean = true);
     function Level1(var _var_:TScData):boolean;
     function Level2(var _var_:TScData):boolean; // &
     function Level3(var _var_:TScData):boolean; // or, and
     function Level4(var _var_:TScData):boolean; // =,<,>,<=,>=
     function Level5(var _var_:TScData):boolean; // - +
     function Level6(var _var_:TScData):boolean;

     procedure FuncLexem(const fname:string; Args:TArgs);
     function EndLexem:boolean;
     procedure IncludeLexem;
     procedure SectionLexem;
     procedure VarLexem;
     procedure GVarLexem;
     procedure ReturnLexem;
     procedure ForLexem;
     procedure WhileLexem;
     procedure IfLexem;
     procedure ElseIfLexem;
     procedure ElseLexem;
     procedure EventLexem;
     procedure PrintLexem;
     procedure PrintLnLexem;
     procedure IncLevelLexem;
     procedure DecLevelLexem;
     procedure VarOpLexem(var_int:PScData);
     procedure ReadCustomValue(var value:TScData);

     function getInternalVar(const name:string; var int_var:PScData):boolean;
     function getGlobalVar(const name:string; var gbl_var:PScData):boolean;
     function isFunction(const name:string; var ind:integer):boolean;
     function isInternalFunction(const name:string; var ind:integer):boolean;
     function isObject(const name:string; var ind:integer):boolean;

     function LinkedLexem:boolean;
     function IsDefLexem:boolean;
     function IsSetLexem:boolean;
     function StrLexem:TScData;
     function CodeLexem:TScData;
     function IsSecLexem:boolean;
     function CallFunc(ind:integer):TScData;
     function CallIntFunc(ind:integer):TScData;
     function CallObject(ind:integer):TScData;

     procedure CallEvent(const event:string; const data:TScData);
    public
     constructor Create;
     destructor Destroy; override;
     function GetToken:boolean;
     procedure PutToken;
     procedure Parse(_cgt:PCodeGenTools; e:id_element; const fname:string; _codeb:TCodeBlock; Args:TArgs; _level:integer; ret_data:PScData);
  end;

 TMapProc = function (cgt:PCodeGenTools; e:id_element; args:TArgs):TScData;
 TObjProc = function (cgt:PCodeGenTools; e:id_element; obj:pointer; index:integer; args:TArgs):TScData;
 TFuncMap = record
   name:string;
   count:integer;   //  -1    
   proc:TMapProc;
 end;
 TObjField = record
   name:string;
 end;
 TObjMethod = record
   name:string;
   count:integer;   //  -1    
 end;
 TObjMap = record
   name:string;
   obj:pointer;
   fields_count:integer;   //   
   methods_count:integer;  //   
   fields:array of TObjField;
   methods:array of TObjMethod;
   method_proc:TObjProc;
 end;

const
  //  ,   GetTok
  TokName   = 1;
  TokNumber = 2;
  TokReal   = 3;
  TokString = 4;
  TokSymbol = 5;
  TokMath   = 6;
  TokCode   = 7;

  //   
  //sec_php = 1;
  //sec_html = 2;
  sec_cpp = 4;

  //   
  //0   - function
  //1-2 - if else
  //3   - while

var el_cnt:integer;
    nsection:integer;    //  
    GVarList:TArgs;      //   

//*******************************************************

function map_replace(cgt:PCodeGenTools; e:id_element; args:TArgs):TScData;
var s:string;
begin
//  cgt._Debug(PChar('Replace...'),clred);
  s := args.Values[0].toCode();
  Replace(s, args.Values[1].toCode(), args.Values[2].toCode());
  args.Values[0].SetValue(s,data_code);
  Result := args.Values[0]^;
end;

function map_trace(cgt:PCodeGenTools; e:id_element; args:TArgs):TScData;
begin
  cgt._Debug(PChar(Args.Values[0].toStr()), clGreen);
  Result.SetValue('');
end;

function map_fopen(cgt:PCodeGenTools; e:id_element; args:TArgs):TScData;
var fs:PStream;
    md,fname:string;
begin
    fname := args.Values[0].toCode();
    md := args.Values[1].toCode();
    if md = 'r' then
      fs := NewReadFileStream(fname)
    else fs := NewWriteFileStream(fname);
    Result.SetValue(integer(fs));
end;

function map_fputs(cgt:PCodeGenTools; e:id_element; args:TArgs):TScData;
var fs:PStream;
begin
  fs := PStream(args.Values[0].toInt());
  fs.WriteStr(args.Values[1].toCode() + #13#10);
  Result.SetValue('');
end;

function map_fgets(cgt:PCodeGenTools; e:id_element; args:TArgs):TScData;
var fs:PStream;
    s:string;
    b:char;
begin
   fs := PStream(args.Values[0].toInt());
   s := '';
   while fs.Position < fs.Size do
     begin
      fs.Read(b,1);
      if b = #13 then
       begin
         fs.Read(b,1);
         break;
       end
      else s := s + b;
     end;
  Result.SetValue(s);
end;

function map_fclose(cgt:PCodeGenTools; e:id_element; args:TArgs):TScData;
var fs:PStream;
begin
   fs := PStream(args.Values[0].toInt());
   fs.Free;
   Result.SetValue('');
end;

function map_project_dir(cgt:PCodeGenTools; e:id_element; args:TArgs):TScData;
var ppath:array[0..255] of char;
begin
//   cgt._Debug(PChar('Element ' + int2hex(e,6)),clred);
   integer(pointer(@ppath)^) := e;
   cgt.GetParam(PARAM_PROJECT_PATH, @ppath);
   Result.SetValue(PChar(@ppath));
end;

const
  //   - 
  func_map_size = 7;
  func_map:array[0..func_map_size-1] of TFuncMap = (
    (name: 'replace';     count:3;  proc:map_replace),
    (name: 'trace';       count:1;  proc:map_trace),
    (name: 'fopen';       count:2;  proc:map_fopen),
    (name: 'fputs';       count:2;  proc:map_fputs),
    (name: 'fgets';       count:1;  proc:map_fgets),
    (name: 'fclose';      count:1;  proc:map_fclose),
    (name: 'project_dir'; count:0;  proc:map_project_dir)
  );
var
  //    
  obj_map_size:integer;
  obj_map:array of TObjMap;

function MakeMethod(const name:string; count:integer):TObjMethod;
begin
   Result.name := name;
   Result.count := count;
end;

function block_proc(cgt:PCodeGenTools; e:id_element; obj:TCodeBlock; index:integer; args:TArgs):TScData;
begin
   Result.SetValue('');
   case index of
    0: obj.Reg(args.Values[0].toStr());
    1: Result.SetValue(obj.Select(args.Values[0].toStr()));
    2: obj.Delete(args.Values[0].toStr());
    3: obj.CopyTo(obj.CurBlockName,args.Values[0].toStr());
    4: Result.SetValue(obj.AsText());
    5: obj.SaveToFile(args.Values[0].toStr(),obj.CurBlockName);
    6: Result.SetValue(obj.CurBlockName);
    7: Result.SetValue(obj.RegGen());
    8: Result.SetValue(obj.empty());
   end;
end;

function cgt_proc(cgt:PCodeGenTools; e:id_element; obj:pointer; index:integer; args:TArgs):TScData;
begin
   case index of
    0: Result.SetValue(cgt.ptGetParent(args.Values[0].toInt()));
    1: Result.SetValue(cgt.ptGetLinkPoint(args.Values[0].toInt()));
    2: Result.SetValue(cgt.elGetClassName(args.Values[0].toInt()));
    3: Result.SetValue(cgt.elGetPtName(e,PChar(args.Values[0].toStr())));
   end;
end;

procedure CreateObjMap;
begin
  obj_map_size := 2;
  SetLength(obj_map, obj_map_size);
  with obj_map[0] do
   begin
     name := 'block';
     obj := TCodeBlock.Create;
     fields_count := 0;
     methods_count := 9;
     SetLength(fields,fields_count);
     SetLength(methods,methods_count);
     methods[0] := MakeMethod('reg',    1);
     methods[1] := MakeMethod('select', 1);
     methods[2] := MakeMethod('delete', 1);
     methods[3] := MakeMethod('copyto', 1);
     methods[4] := MakeMethod('astext', 0);
     methods[5] := MakeMethod('save',   1);
     methods[6] := MakeMethod('cur',    0);
     methods[7] := MakeMethod('reggen', 0);
     methods[8] := MakeMethod('empty',  0);
     method_proc := TObjProc(@block_proc);
   end;
  with obj_map[1] do
   begin
     name := 'cgt';
     obj := nil;
     fields_count := 0;
     methods_count := 4;
     SetLength(fields,fields_count);
     SetLength(methods,methods_count);
     methods[0] := MakeMethod('pt_get_parent',     1);
     methods[1] := MakeMethod('pt_get_link_point', 1);
     methods[2] := MakeMethod('el_get_class_name', 1);
     methods[3] := MakeMethod('get_point_id',      1);
     method_proc := cgt_proc;
   end;
end;

procedure DestroyObjMap;
begin
  TCodeBlock(obj_map[0].Obj).Destroy;
end;

//*******************************************************

procedure TScData.SetValue(const val:string);
begin
   Value.data_type := data_str;
   Value.sdata := val;
end;

procedure TScData.SetValue(val:integer);
begin
   Value.data_type := data_int;
   Value.idata := val;
end;

procedure TScData.SetValue(val:real);
begin
   Value.data_type := data_real;
   Value.rdata := val;
end;

procedure TScData.SetValue(val:boolean);
begin
   Value.data_type := data_int;
   Value.idata := integer(val);
end;

procedure TScData.SetValue(const val:TData);
begin
   Value := val;
end;

procedure TScData.SetValue(const val:string; tp:byte);
begin
   Value.data_type := tp;
   Value.sdata := val;
end;

procedure TScData.AddValue;
var arr:PScArray;
    i:integer;
begin
   if Value.data_type <> data_array then
    begin
      new(arr);
      FillChar(arr^,sizeof(TScArray),0);
      inc(arr.count);
      SetLength(arr.items,arr.count);
      arr.items[arr.count-1].SetValue(value);
      Value.idata := integer(arr);
      Value.data_type := data_array;
    end
   else arr := PScArray(value.idata);

   if val.GetType = data_array then
    begin
     SetLength(arr.items,arr.count + val.toArray.Count);
     for i := 0 to val.toArray.Count-1 do
       arr.items[arr.count + i] := val.toArray.Items[i];
     inc(arr.count,val.toArray.Count);
    end
   else
    begin
     inc(arr.count);
     SetLength(arr.items,arr.count);
     arr.items[arr.count-1] := val;
    end;
end;

function TScData.GetType;
begin
   Result := Value.data_type;
end;

procedure TScData.SetType;
begin
   Value.data_type := tp;
end;

procedure TScData.GetValue;
begin
   val := value;
end;

function TScData.toStr;
var i:integer;
    arr:PScArray;
begin
   case Value.data_type of
    data_int: Result := int2str(Value.idata);
    data_str,data_code: Result := Value.sdata;
    data_real: Result := Double2str(Value.rdata);
    data_array:
     begin
      arr := PScArray(value.idata);
      Result := '';
      for i := 0 to arr.count-1 do
        Result := Result + arr.items[i].toStr;
     end;
    data_null: Result := '';
   end;
end;

function TScData.toInt;
begin
   case Value.data_type of
    data_int: Result := Value.idata;
    data_str: Result := str2int(Value.sdata);
    data_real: Result := Round(Value.rdata);
    else Result := 0;
   end;
end;

function TScData.toReal;
begin
   case Value.data_type of
    data_int: Result := Value.idata;
    data_str: Result := str2double(Value.sdata);
    data_real: Result := Value.rdata;
    else Result := 0;
   end;
end;

function TScData.toCode;
var i:integer;
    arr:PScArray;
begin
//   case nsection of
//    sec_cpp:
       case Value.data_type of
        data_int: Result := int2str(Value.idata);
        data_str: Result := '"' + Value.sdata + '"';
        data_real: Result := double2str(Value.rdata);
        data_code: Result := Value.sdata;
        data_array:
         begin
          arr := PScArray(value.idata);
          if arr.items[0].GetType = data_code then
            Result := ''
          else Result := '"';
          Result := Result + arr.items[0].toStr;

          for i := 1 to arr.count-1 do
            if(arr.items[i-1].GetType = data_code) then
              if arr.items[i].GetType = data_code then
                Result := Result + ' + ' + arr.items[i].toStr
              else Result := Result + ' + "' + arr.items[i].toStr
            else if(arr.items[i].GetType = data_code)then
              if arr.items[i-1].GetType = data_code then
                Result := Result + ' + ' + arr.items[i].toStr
              else Result := Result + '" + ' + arr.items[i].toStr
            else Result := Result + arr.items[i].toStr;
          if arr.items[arr.count-1].GetType = data_code then

          else Result := Result + '"';
         end;
        else Result := '""';
       end;
//    else Result := '';
//   end;
end;

function TScData.isEmpty;
begin
   case Value.data_type of
    data_int: Result := Value.idata = 0;
    data_str,data_code: Result := Value.sdata = '';
    data_real: Result := Value.rdata = 0;
    data_array: Result := false;
    else Result := true;
   end;
end;

function TScData.toArray;
begin
   Result := PScArray(value.idata);
end;

//____________________

function StrIComp(const Str1, Str2: PChar): Integer; assembler;
asm
        PUSH    EDI
        PUSH    ESI
        MOV     EDI,EDX
        MOV     ESI,EAX
        MOV     ECX,0FFFFFFFFH
        XOR     EAX,EAX
        REPNE   SCASB
        NOT     ECX
        MOV     EDI,EDX
        XOR     EDX,EDX
@@1:    REPE    CMPSB
        JE      @@4
        MOV     AL,[ESI-1]
        CMP     AL,'a'
        JB      @@2
        CMP     AL,'z'
        JA      @@2
        SUB     AL,20H
@@2:    MOV     DL,[EDI-1]
        CMP     DL,'a'
        JB      @@3
        CMP     DL,'z'
        JA      @@3
        SUB     DL,20H
@@3:    SUB     EAX,EDX
        JE      @@1
@@4:    POP     ESI
        POP     EDI
end;

function MakeSpaces(level:integer):string;
var i:integer;
begin
  Result := '';
  for i := 1 to level do
    Result := Result + ' ';
end;

function findProperty(cgt:PCodeGenTools; el:id_element; name:PChar):integer;
var i:integer;
begin
   Result := -1;
   for i := 0 to cgt.elGetPropCount(el)-1 do
     if strcomp(cgt.propGetName(cgt.elGetProperty(el, i)), name) = 0 then
      begin
        Result := i;
        Exit;
      end
end;

function DoData(cgt:PCodeGenTools; dt:id_data):TScData;
begin
     case cgt.dtType(Dt) of
       data_null: Result.SetValue('');
       data_int:  Result.SetValue(cgt.dtInt(Dt));
       data_str:  Result.SetValue(cgt.dtStr(Dt));
       data_real: Result.SetValue(cgt.dtReal(Dt));
     end;
end;

function readProperty(cgt:PCodeGenTools; prop:id_prop):TScData;
begin
  case cgt.propGetType(prop) of
    data_int: Result.SetValue(integer(cgt.propGetValue(prop)^));
    data_str: Result.SetValue(string(cgt.propGetValue(prop)^));
    data_real: Result.SetValue(real(cgt.propGetValue(prop)^));
    data_data: Result := DoData(cgt,id_data(cgt.propGetValue(prop)));
    data_combo: Result.SetValue(byte(cgt.propGetValue(prop)^));
    data_list: Result.SetValue(cgt.propToString(prop));
    data_comboEx: Result.SetValue(cgt.propToString(prop), data_code);
    else Result.SetValue('');
  end;
end;

function CreateCode(SDK:cardinal; cgt:PCodeGenTools):PStrList; cdecl;
var
    i:integer;
    e:cardinal;
    Parser:TParser;
    codeb:TCodeBlock;
    Args:TArgs;
begin
   CreateObjMap;

   Result := NewStrList;
   codeb := TCodeBlock(obj_map[0].obj);

   el_cnt := 0;
   Parser := TParser.Create;
   GVarList := TArgs.Create;
   for i := 0 to cgt.sdkGetCount(sdk)-1 do
    begin
      e := cgt.sdkGetElement(sdk,i);
      cgt.elSetCodeName(e,'');
    end;
   for i := 0 to cgt.sdkGetCount(sdk)-1 do
    begin
     e := cgt.sdkGetElement(sdk,i);
     if pos('MainForm',cgt.elGetClassName(e)) > 0 then
       begin
         //if cgt.elGetClassName(e) = 'EntryPoint' then
         //  nsection := sec_php
         //else if cgt.elGetClassName(e) = 'JavaEntryPoint' then
         nsection := sec_cpp;
         //else nsection := sec_html;
         Args := TArgs.Create;
         Parser.Parse(cgt, e, 'dostart', codeb, Args, 0, nil);
         Args.Destroy;
         break;
       end;
    end;
   codeb.Merge(Result);
   Parser.Destroy;
   GVarList.Destroy;

   DestroyObjMap;
end;

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
procedure TBlockStack.Push;
begin
   inc(Count);
   SetLength(Items,Count);
   Items[Count-1].skip := _skip;
   Items[Count-1].LType := _Type;
   Items[Count-1].userdata := _userdata;
end;

function TBlockStack.Pop;
begin
   if Count > 0 then
    begin
     _Skip := Items[Count-1].skip;
     _Type := Items[Count-1].LType;
     _userdata := Items[Count-1].userdata;
     dec(Count);
     Result := false;
    end
   else Result := true;
end;

procedure TBlockStack.Clear;
begin
   Count := 0;
   SetLength(Items,Count);
end;

//**************************************************************************************

constructor TCodeBlock.Create;
begin
   inherited;
   FItems := NewStrListEx;
end;

destructor TCodeBlock.Destroy;
var i:integer;
begin
   for i := 0 to Count-1 do
    Items[i].Free;
   inherited;
end;

procedure TCodeBlock.Reg;
begin
   if Find(name) <> nil then exit;

   FItems.AddObject(LowerCase(name), cardinal(NewStrList));
end;

function TCodeBlock.RegGen;
begin
   Result := 'block_' + int2str(FGen);
   inc(FGen);
   Reg(Result);
end;

procedure TCodeBlock.Select;
var l:PStrList;
begin
   l := Find(name);
   if l <> nil then
     begin
      Current := l;
      Result := FCurName;
      FCurName := name;
     end
   else Result := '';
end;

procedure TCodeBlock.Delete;
var i:integer;
begin
   for i := 0 to Count-1 do
    if StrIComp(PChar(FItems.Items[i]), PChar(name)) = 0 then
     begin
       if Current = Items[i] then
        Current := nil;
       Items[i].Free;
       FItems.Delete(i);
       break;
     end;
end;

procedure TCodeBlock.CopyTo;
var src,dst:PStrList;
begin
    src := Find(Source);
    dst := Find(Dest);
    if(src <> nil)and(dst <> nil)then
      dst.AddStrings(src);
end;

procedure TCodeBlock.Merge;
var i:integer;
begin
   for i := 0 to Count-1 do
    List.AddStrings(Items[i]);
end;

procedure TCodeBlock.Add;
begin
   if Current <> nil then
     Current.Add(Text);
end;

procedure TCodeBlock.Paste;
begin
   if Current <> nil then
    begin
      if Current.Count = 0 then
        Current.Add(text)
      else  Current.Items[Current.Count-1] := Current.Items[Current.Count-1] + Text;
    end;
end;

procedure TCodeBlock.AddStrings;
begin
   if Current <> nil then
     Current.AddStrings(list);
end;

function TCodeBlock.Empty;
var i:integer;
begin
   Result := true;
   for i := 0 to Count-1 do
    Result := Result and (Items[i].Count = 0);
end;

function TCodeBlock.AsText;
begin
   Result := Current.Text;
end;

procedure TCodeBlock.SaveToFile;
var lst:PStrList;
begin
   lst := Find(name);
   if lst <> nil then
    lst.SaveToFile(fname);
end;

function TCodeBlock.Find;
var i:integer;
begin
   for i := 0 to Count-1 do
    if StrIComp(PChar(FItems.Items[i]), PChar(name)) = 0 then
     begin
       Result := Items[i];
       exit;
     end;
   Result := nil;
end;

function TCodeBlock.GetItems;
begin
   if(Index >= 0)and(Index < Count)then
     Result := PStrList(FItems.Objects[Index])
   else Result := nil;
end;

function TCodeBlock.GetCount;
begin
  Result := FItems.Count;
end;

//**************************************************************************************

constructor TArgs.Create;
begin
   inherited;
   FItems := NewStrListEx;
end;

destructor TArgs.Destroy;
var i:integer;
begin
   for i := 0 to Count-1 do
    begin
     Values[i].value.sdata := '';
     dispose(Values[i]);
    end;
   FItems.Free;
   inherited;
end;

function TArgs.GetValues;
begin
   if(index >= 0)and(index < Count)then
     Result := PScData(FItems.objects[index])
   else Result := nil;
end;

function TArgs.GetNames;
begin
   if(index >= 0)and(index < Count)then
     Result := FItems.items[index]
   else Result := '';
end;

procedure TArgs.SetNames;
begin
   if(index >= 0)and(index < Count)then
     FItems.items[index] := lowercase(value);
end;

function TArgs.GetCount;
begin
   Result := FItems.Count;
end;

function TArgs.find;
var i:integer;
begin
   for i := 0 to COunt-1 do
    if name = names[i] then
     begin
       Result := Values[i];
       exit;
     end;
   Result := nil;
end;

function TArgs.AddArg;
begin
   new(Result);
   FillChar(Result^,sizeof(TScData),0);
   FItems.AddObject(lowercase(Name), cardinal(Result));
end;

//**************************************************************************************

constructor TParser.Create;
begin
   inherited Create;
   BlockStack := TBlockStack.Create;
   Lines := NewStrList;
   VarList := TArgs.Create;
   funcList := NewStrListEx;
   fData := nil;
end;

destructor TParser.Destroy;
var i:integer;
begin
   BlockStack.Destroy;
   Lines.Free;
   VarList.Destroy;
   for i := 0 to funcList.Count-1 do
    dispose(PFuncData(funcList.Objects[i]));
   funcList.Free;
   inherited;
end;

procedure TParser.Debug;
begin
   {$ifdef debug}
   cgt._Debug(PChar(Text), color);
   {$endif}
end;

function TParser.isDebug;
begin
   Result := true; //cgt.isDebug(cgt.elGetSDK(el));
end;

function TParser.ReadLine;
begin
   while LineIndex < Lines.Count do
    begin
       Line := Lines.Items[LineIndex] + #1;
       inc(LineIndex);
       if Line <> #1 then break;
    end;
   Result := LineIndex = Lines.Count;
end;

procedure TParser.AddError;
begin
   cgt._Debug(PChar('Error ' + uname + '[' + int2str(LineIndex) + ']: ' + Error), clRed);
   ReadLine;
   LPos := 1;
   _err_ := true;
end;

procedure TParser.PutToken;
begin
   LineIndex := oldLineIndex;
   Line := Lines.Items[lineindex-1] + #1;
   Lpos := oldLPos;
   Token := '';
end;

function TParser.GetToken;
var err:string;
begin
   oldLineIndex := LineIndex;
   oldLPos := LPos;
   Token := '';
   RToken := '';
   TokType := 0;
   err := '';
   Result := false;
   repeat
    case Line[LPos] of
     ' ',#9: inc(LPos);
     'a'..'z','A'..'Z','_',''..'',''..'':
       begin
         repeat
           RToken := RToken + Line[LPos];
           inc(LPos);
         until not(Line[LPos] in ['a'..'z','A'..'Z','_',''..'',''..'','0'..'9']);
         Token := LowerCase(RToken);
         TokType := TokName;
       end;
     '0'..'9':
       begin
         repeat
           Token := Token + Line[LPos];
           inc(LPos);
         until not(Line[LPos] in ['0'..'9']);
         if (Line[LPos] = '.')and(Line[LPos+1] <> '.') then
          begin
           repeat
            Token := Token + Line[LPos];
            inc(LPos);
           until not(Line[LPos] in ['0'..'9']);
           if Token[Length(Token)] = '.' then
             AddError('Syntax error. Please write correct real number.');
           TokType := TokReal;
          end
         else TokType := TokNumber;
       end;
     '"':
       begin
         inc(LPos);
         while (Line[LPos] <> #1) do
          begin
           case Line[LPos] of
            '"': break;
            '\':
              if Line[LPos+1] in ['\','"'] then
                 inc(LPos);
           end;
           Token := Token + Line[LPos];
           inc(LPos);
          end;
         TokType := TokString;
         if Line[LPos] = #1 then
           err := Format(ER_LEX_NOT_FOUND, [' " '])
         else inc(LPos);
       end;
     '''':
       begin
         inc(LPos);
         while (Line[LPos] <> #1) do
          begin
           case Line[LPos] of
            '''': break;
            '\':
              if Line[LPos+1] in ['\',''''] then
                 inc(LPos);
           end;
           Token := Token + Line[LPos];
           inc(LPos);
          end;
         TokType := TokCode;
         if Line[LPos] = #1 then
           err := Format(ER_LEX_NOT_FOUND, [' '' '])
         else inc(LPos);
       end;
     '(',')','=',',',':','.':
        begin Token := Line[LPos]; inc(LPos); TokType := TokSymbol; end;
     '+','-','/','*','&':
        begin
          Token := Line[LPos];
          inc(LPos);
          if Line[LPos] = '/' then
            while Line[LPos] <> #1 do inc(LPos)
          else if Line[LPos] = '=' then
            begin
               Token := Token + Line[LPos];
               inc(LPos);
               TokType := TokSymbol
            end
          else TokType := TokMath;
        end;
     '>','<','!':
        begin
          Token := Line[LPos];
          inc(LPos);
            if Line[LPos] = '=' then
             begin
                Token := Token + Line[LPos];
                inc(LPos);
             end
            else if ( Token = '<' )and(Line[LPos] = '>')then
             begin
                Token := '!=';
                inc(LPos);
             end;
          TokType := TokMath;
        end;
     '#':
        begin
          TokType := TokString;
          inc(LPos);
          Token := '';
          while (Line[LPos] in ['0'..'9'])and(Line[LPos] <> #1) do
           begin
             Token := Token + Line[LPos];
             inc(LPos);
           end;
          if Token = '' then
           Token := #0
          else Token := chr(str2int(Token));
        end;
     #1: begin ReadLine; LPos := 1; end;
     #2: Result := true;
     else
      begin
       err := 'Symbol ' + Line[LPos] + ' don''t support';
       inc(LPos);
      end;
    end;
   until (TokType <> 0)or(err <> '')or(Line[LPos] = #2);

   if err <> '' then
    begin
     Result := true;
     AddError(err);
    end;
end;

function TParser.ReadPoint(lp:id_point):TScData;
var
    Parser:TParser;
    ret:TScData;
    fname:string;
    Args:TArgs;
    el:id_element;
begin
  Parser := TParser.Create;
  el := cgt.ptGetParent(lp);
  Args := TArgs.Create;
  if (cgt.elGetClassIndex(el) = CI_DPElement)and(cgt.ptGetType(lp) = pt_var)and(length(cgt.pt_dpeGetName(lp)) > 0)then
    begin
     fname := lowercase(cgt.pt_dpeGetName(lp));
     args.AddArg('').SetValue(cgt.ptGetIndex(lp));
    end
  else fname := lowercase(cgt.ptGetName(lp));
  Parser.Parse(cgt, el, fname, codeb, args, 1, @ret);
  Result := ret;
  Args.Destroy;
  Parser.Destroy;
end;

procedure TParser.Parse;
var UnitName{,old}:string;
    _args:TArgs;
    Parser:TParser;
begin
   cgt := _cgt;
   el := e;
   level := _level;
   codeb := _codeb;

   section := sec_cpp;

   if Length(cgt.elGetCodeName(el)) = 0 then
    begin
     cgt.elSetCodeName(el, PChar(Int2Str(el_cnt)));
     inc(el_cnt);

     Parser := TParser.Create;
     _Args := TArgs.Create;
     {$ifdef init_section}
     Debug('Call init section for ' + cgt.elGetClassName(e));
     {$endif}
     Parser.Parse(cgt,el,'init',codeb,_Args,1,nil);
     _Args.Destroy;
     Parser.Destroy;
    end;

   uname := 'hi' + cgt.elGetClassName(e) + '.hws';
   UnitName := cgt.ReadStrParam(PARAM_CODE_PATH) + uname;
   Lines.LoadFromFile(UnitName);
   Lines.Add(#2);
   {$ifdef processed_unit}
   Debug('Load unit: ' + uname);
   {$endif}

   if cgt.elGetClassIndex(e) = CI_DPElement then
     ;

   LineIndex := 0;
   Line := '';
   LPos := 1;
   Start(fname, args);
   if ret_data <> nil then
    ret_data^ := _data_return_;
end;

procedure TParser.FuncLexem;
var sec,i:integer;
begin
   if GetToken then exit;

   {$ifdef processed_func}
   Debug('Process func: ' + RToken);
   {$endif}

   funcList.Add(token);
   new(fData);
   FillChar(fData^,sizeof(TFuncData),0);
   funcList.Objects[funcList.Count-1] := cardinal(fdata);
   fData.Line := LineIndex;
   fData.Skeep := false;
   //\nDebug('Add function to list: ' + RToken + ':' + fname);
   fData._args_ := TArgs.Create;
   fData._data_saved_.SetValue('');

   fData.fSkeep := not (fname = Token);
   {$ifdef called_func}
   if not fData.fSkeep then
     Debug('Called func: ' + RToken);
   {$endif}

   fData.fsection := section;
   if GetToken then exit;
   if Token = '(' then
    begin
      if GetToken then exit;
      while token <> ')' do
       begin
         if token = ',' then
          if GetToken then exit;

         if toktype <> tokname then
          begin
            AddError('Name expected');
            Exit;
          end;

         fData._args_.AddArg(token);

         if GetToken then exit;
         if(token <> ',')and(token <> ')')then
          begin
            AddError('Syntax error for arguments');
            Exit;
          end;
       end;
      if GetToken then exit;
    end;

   if Token = ':' then
    begin
     repeat
      if GetToken then exit;
      if ToSection(sec) then exit;
      fData.fSection := fData.fSection or sec;
      if GetToken then exit;
     until token <> ',';
    end;
   PutToken;

   //\nDebug('Need section: ' + int2str(fsection));

   fData.fSkeep := fData.fSkeep or (fData.fsection and nsection = 0);
   if not fData.fSkeep then
    begin
      {$ifdef called_func_sec}
      Debug(' Called func: ' + RToken + '(' + int2str(fData._args_.Count) + ') for section ' + int2str(nsection));
      {$endif}
      for i := 0 to min(args.Count,fData._args_.Count)-1 do
       fData._args_.values[i]^ := args.Values[i]^;
      if args.count > 0 then
         fData._data_saved_ := args.Values[0]^;
    end;

   fData.Skeep := false;
   BlockStack.Push(fData.fSkeep, 0, nil);
end;

procedure TParser.EndLexem;
var
    tp:word;
    p:pointer;
begin
  Result := false;
  if BlockStack.Count = 0 then
    begin
      AddError('Syntax error: END lexem');
      Exit;
    end;

  BlockStack.Pop(fData.Skeep, tp, p);
  case tp of
   0:
    begin
     fData.fSkeep := fData.Skeep;
     Result := not fData.fSkeep;
     fData.Skeep := false;
    end;
   3:
    if p <> nil then
     begin
      LineIndex := integer(p);
      LPos := 1;
      ReadLine;
     end;
  end;
end;

procedure TParser.IncludeLexem;
var fname:TScData;
    s:string;
    procedure InsertStrings;
    var list:PStrList;
        i:integer;
    begin
       list := NewStrList;
       list.LoadFromFile(s);
       for i := 0 to list.count-1 do
         Lines.Insert(LineIndex+i, list.items[i]);
       list.Free;
    end;
begin
   if (fData <> nil) and (fData.fSkeep or fData.Skeep) then
    begin
      ReadLine;
      LPos := 1;
      exit;
    end;

   if GetToken then exit;
   if Token <> '(' then
     begin
       AddError('Syntax error for include: (');
       Exit;
     end;
   if level1(fname) then exit;
   s := cgt.ReadStrParam(PARAM_CODE_PATH) + fname.tocode + '.hws';
   if FileExists(s) then
     InsertStrings
   else
     begin
       AddError('File ' + s + ' not found');
       Exit;
     end;

   if GetToken then exit;
   if Token <> ')' then
     begin
       AddError('Syntax error for include: )');
       Exit;
     end;
end;

function TParser.ToSection;
begin
   Result := false;
   //if token = 'php' then
   // sec := sec_php
   //else if token = 'java' then
    sec := sec_cpp
   //else if token = 'html' then
   // sec := sec_html
   //else
   // begin
   //   AddError('Unknown section ' + RToken);
   //   Result := true;
   // end;
end;

procedure TParser.SaveState;
begin
  state.Line := Line;
  state.LineIndex := LineIndex;
  state.LPos := LPos;
  state.Token := Token;
  state.RToken := RToken;
  state.TokType := TokType;
  state.fData := fData;
end;

procedure TParser.LoadState;
begin
  Line := state.Line;
  LineIndex := state.LineIndex;
  LPos := state.LPos;
  Token := state.Token;
  RToken := state.RToken;
  TokType := state.TokType;
  fData := state.fData;
end;

function TParser._points_count_;
var i:integer;
begin
   Result := 0;
   for i := 0 to cgt.elGetPtCount(el)-1 do
    if cgt.ptGetType(cgt.elGetPt(el,i)) = ptype then
      inc(Result);
end;

procedure TParser.SectionLexem;
begin
   if GetToken then exit;
   ToSection(section);
end;

procedure TParser.VarLexem;
{var value:PScData;}
begin
   {
   if fSkeep or Skeep then
    begin
      ReadLine;
      LPos := 1;
      exit;
    end;
   }
   if GetToken then exit;
   if Token <> '(' then
     begin
       AddError('Syntax error for var: (');
       Exit;
     end;

   if GetToken then exit;
   repeat
     VarList.AddArg(Token).SetValue('');
     if GetToken then exit;

     if token = ',' then
       if GetToken then exit;
   until (Token = ')')or(toktype <> tokname);
   if TokType <> TokSymbol then
     begin
       AddError('Syntax error for var: )');
       Exit;
     end;
end;

procedure TParser.GVarLexem;
{var value:PScData;}
begin

   if fData.fSkeep or fData.Skeep then
    begin
      ReadLine;
      LPos := 1;
      exit;
    end;

   if GetToken then exit;
   if Token <> '(' then
     begin
       AddError('Syntax error for var: (');
       Exit;
     end;

   if GetToken then exit;
   repeat
     if GVarList.find(Token) = nil then
       GVarList.AddArg(Token).SetValue('');
     if GetToken then exit;

     if token = ',' then
       if GetToken then exit;
   until (Token = ')')or(toktype <> tokname);
   if TokType <> TokSymbol then
     begin
       AddError('Syntax error for var: )');
       Exit;
     end;
end;

procedure TParser.IncLevelLexem;
begin
   if fData.fSkeep or fData.Skeep then
    begin
      ReadLine;
      LPos := 1;
      exit;
    end;

   if GetToken then exit;
   if Token <> '(' then
     begin
       AddError('Syntax error for var: (');
       Exit;
     end;

   inc(level);

   if GetToken then exit;
   if TokType <> TokSymbol then
     begin
       AddError('Syntax error for var: )');
       Exit;
     end;
end;

procedure TParser.DecLevelLexem;
begin
   if fData.fSkeep or fData.Skeep then
    begin
      ReadLine;
      LPos := 1;
      exit;
    end;

   if GetToken then exit;
   if Token <> '(' then
     begin
       AddError('Syntax error for var: (');
       Exit;
     end;

   dec(level);
   if GetToken then exit;
   if TokType <> TokSymbol then
     begin
       AddError('Syntax error for var: )');
       Exit;
     end;
end;

procedure TParser.ReturnLexem;
var data:TScData;
begin
   if fData.fSkeep or fData.Skeep then
    begin
      ReadLine;
      LPos := 1;
      exit;
    end;

   if GetToken then exit;
   if Token <> '(' then
     begin
       AddError('Syntax error for return: (');
       Exit;
     end;
   _data_return_.SetValue('');
   if Level1(data) then exit;
   _data_return_ := data;

   if GetToken then exit;
   if Token <> ')' then
     begin
       AddError('Syntax error for return: )');
       Exit;
     end;
end;

procedure TParser.ForLexem;
{var data:TScData;}
begin
   if fData.fSkeep or fData.Skeep then
    begin
      ReadLine;
      LPos := 1;
      exit;
    end;

   if GetToken then exit;
   if Token <> '(' then
     begin
       AddError('Syntax error for: (');
       Exit;
     end;

   {
    for(i; i < _event_count_; i++)
      event(onEvent & i, _data_)
    end
   }


   if GetToken then exit;
   if Token <> ')' then
     begin
       AddError('Syntax error for: )');
       Exit;
     end;
end;

procedure TParser.WhileLexem;
var data:TScData;
begin
   if fData.fSkeep or fData.Skeep then
    begin
      BlockStack.Push(false, 3, nil);
      ReadLine;
      LPos := 1;
      exit;
    end;

   if GetToken then exit;
   if Token <> '(' then
     begin
       AddError('Syntax error for while: (');
       Exit;
     end;

   if level1(data) then exit;
   fData.Skeep := data.toInt = 0;
   if fData.Skeep then
     BlockStack.Push(false, 3, nil)
   else BlockStack.Push(false, 3, pointer(lineindex-1));

   if GetToken then exit;
   if Token <> ')' then
     begin
       AddError('Syntax error for while: )');
       Exit;
     end;
end;

procedure TParser.IfLexem;
var tp:word;
    sk:boolean;
    data:TScData;
begin
   if fData.fSkeep or fData.Skeep then
    begin
      ReadLine;
      LPos := 1;
      BlockStack.Push(true, 1, nil);
      exit;
    end;

   if GetToken then exit;
   if Token <> '(' then
     begin
       AddError('Syntax error for if: (');
       Exit;
     end;
   if Level1(data) then exit;
   sk := fData.Skeep;
   fData.Skeep := fData.Skeep or (data.toInt = 0);
   if fData.Skeep then tp := 1 else tp := 2;
   BlockStack.Push(Sk, tp, nil);

   if GetToken then exit;
   if Token <> ')' then
     begin
       AddError(Format(ER_FUNC_ARGS_CLOSE,['if']) + Token);
       Exit;
     end;
end;

procedure TParser.ElseIfLexem;
var _sk:boolean;
    _tp:word;
    data:TScData;
    p:pointer;
begin
   if fData.fSkeep then
    begin
      ReadLine;
      LPos := 1;
      exit;
    end;

   if GetToken then exit;
   if Token <> '(' then
     begin
       AddError('Syntax error for elseif: (');
       Exit;
     end;
   if Level1(data) then exit;
   BlockStack.Pop(_sk, _tp, p);
   if(_tp <> 2) and fData.Skeep and not _sk then
     begin
      fData.Skeep := data.toInt = 0;
      if not fData.Skeep then
       _tp := 2;
     end
   else fData.Skeep := true;
   BlockStack.Push(_sk, _tp, p);

   if GetToken then exit;
   if Token <> ')' then
     begin
       AddError('Syntax error for elseif: )');
       Exit;
     end;
end;

procedure TParser.ElseLexem;
var _sk:boolean;
    _tp:word;
    p:pointer;
begin
   if fData.fSkeep then
    begin
      ReadLine;
      LPos := 1;
      exit;
    end;

   BlockStack.Pop(_sk, _tp, p);
   if (_tp = 1)and(not _sk) then
     fData.Skeep := not fData.Skeep
   else fData.Skeep := true;
   BlockStack.Push(_sk, _tp, p);
end;

procedure TParser.EventLexem;
var ename:string;
    data:TScData;
    oldsec:integer;
begin
   if fData.fSkeep or fData.Skeep then
    begin
      ReadLine;
      LPos := 1;
      exit;
    end;

   if GetToken then exit;
   if Token <> '(' then
     begin
       AddError('Syntax error for event: (');
       Exit;
     end;

   if level1(data) then exit;
   ename := data.toStr;
   oldsec := nsection;
   if GetToken then exit;
   if Token = ':' then
    begin
      if GetToken then exit;
      if ToSection(nsection) then exit;
      Debug('Call event with ' + int2str(nsection) + ' section');
      if GetToken then exit;
    end;

   if Token = ',' then
     begin
       if level1(data) then exit;
       if GetToken then exit;
     end
   else data.SetValue('');
   {$ifdef event_call}
   Debug('Call event with ' + ename);
   {$endif}
   CallEvent(ename, data);
   nsection := oldsec;

   if Token <> ')' then
     begin
       AddError('Syntax error for event: )');
       Exit;
     end;
end;

procedure TParser.PrintLexem;
var s:string;
    data:TScData;
begin
   if fData.fSkeep or fData.Skeep then
    begin
      ReadLine;
      LPos := 1;
      exit;
    end;

   if GetToken then exit;
   if Token <> '(' then
     begin
       AddError('Syntax error: (');
       Exit;
     end;

   s := '';
   if Level1(data) then exit;
   repeat
     if Token = ')' then
       begin
         AddError('Syntax error: no argument found');
         Exit;
       end
     else if token = ',' then
       if Level1(data) then exit;
     s := s + data.toCode;
     if GetToken then exit;
   until Token = ')';
   codeb.Paste(MakeSpaces(level) + s);
   {$ifdef print_lexem}
   Debug('Print[' + codeb.CurBlockName + ']: ' + s);
   {$endif}
end;

procedure TParser.PrintLnLexem;
begin
   if fData.fSkeep or fData.Skeep then
    begin
      ReadLine;
      LPos := 1;
      exit;
    end;

   PrintLexem;
   codeb.Add('');
end;

procedure tparser.VarOpLexem;
var _var_:TScData;
begin
   if fData.fSkeep or fData.Skeep then
    begin
      ReadLine;
      LPos := 1;
      exit;
    end;
   if GetToken then exit;
   if token = '=' then
     begin
       if Level1(_var_) then exit;
       var_int^ := _var_;
     end
   else if token = '&=' then
     begin
       if Level1(_var_) then exit;
       var_int.AddValue(_var_);
     end
   else if token = '+=' then
     begin
       if Level1(_var_) then exit;
       SumVar(var_int^,_var_);
     end
   else
     AddError('Syntax error: uncnown operator');
end;

procedure TParser.ReadCustomValue;
var dt:TScData;
begin
   if GetToken then exit;
   if Token <> '(' then
     begin
       AddError('Syntax error for linked: (');
       Exit;
     end;

   if level1(dt) then exit;
   ReadValue(dt.toStr, value);

   if GetToken then exit;
   if Token <> ')' then
     begin
       AddError('Syntax error for linked: )');
       Exit;
     end;
end;

function TParser.LinkedLexem;
var lp:id_point;
begin
   if GetToken then exit;
   if Token <> '(' then
     begin
       AddError('Syntax error for linked: (');
       Exit;
     end;
   if GetToken then exit;
   lp := cgt.elGetPtName(el, PChar(RToken));
   if lp > 0 then
    begin
     lp := cgt.ptGetRLinkPoint(lp);
     Result := lp > 0;
    end
   else Result := false;

   if GetToken then exit;
   if Token <> ')' then
     begin
       AddError('Syntax error for linked: )');
       Exit;
     end;
end;

function TParser.IsDefLexem;
var pr:id_prop;
begin
   if GetToken then exit;
   if Token <> '(' then
     begin
       AddError('Syntax error for isdef: (');
       Exit;
     end;
   if GetToken then exit;
   pr := findProperty(cgt, el, PChar(RToken));
   if pr <> -1 then
     Result := cgt.elIsDefProp(el, pr)
   else
    begin
     Debug('Property ' + RToken + ' not found!', clRed);
     Result := false;
    end;

   if GetToken then exit;
   if Token <> ')' then
     begin
       AddError('Syntax error for isdef: )');
       Exit;
     end;
end;

function TParser.IsSetLexem;
var _var_:TScData;
begin
   if GetToken then exit;
   if Token <> '(' then
     begin
       AddError('Syntax error for isset: (');
       Exit;
     end;
   if GetToken then exit;

   ReadValue(RToken, _var_, false);
   Result := not _var_.isEmpty;

   if GetToken then exit;
   if Token <> ')' then
     begin
       AddError('Syntax error for isset: )');
       Exit;
     end;
end;

function TParser.IsSecLexem;
var
    sec:integer;
begin
   if GetToken then exit;
   if Token <> '(' then
     begin
       AddError('Syntax error for issec: (');
       Exit;
     end;
   if GetToken then exit;

   if ToSection(sec) then
     Result := false
   else Result := nsection and sec > 0;

   if GetToken then exit;
   if Token <> ')' then
     begin
       AddError('Syntax error for issec: )');
       Exit;
     end;
end;

function TParser.StrLexem;
var data:TScData;
begin
   if GetToken then exit;
   if Token <> '(' then
     begin
       AddError('Syntax error for str: (');
       Exit;
     end;
   if level1(data) then exit;
   Result.SetValue(data.toStr);
   //\nDebug('Convert to str: ' + data.tostr);

   if GetToken then exit;
   if Token <> ')' then
     begin
       AddError('Syntax error for str: )');
       Exit;
     end;
end;

function TParser.CodeLexem;
var data:TScData;
begin
   if GetToken then exit;
   if Token <> '(' then
     begin
       AddError('Syntax error for code: (');
       Exit;
     end;
   if level1(data) then exit;
   Result.SetValue(data.toStr, data_code);
   //\nDebug('Convert to code: ' + data.tostr);

   if GetToken then exit;
   if Token <> ')' then
     begin
       AddError('Syntax error for code: )');
       Exit;
     end;
end;

function TParser.CallFunc;
var
    old_state:TFuncState;
    fname:string;
    i:integer;
    fd:PFuncData;
begin
   {$ifdef call_func}
   Debug('Call user function ' + RToken);
   {$endif}
   fname := Token;

   if GetToken then exit;
   if Token <> '(' then
     begin
       AddError('Syntax error for call func: (');
       Exit;
     end;

   fd := PFuncData(funcList.Objects[ind]);
   for i := 0 to fd._args_.Count-1 do // todo: check errors
    begin
      if i > 0 then
       begin
         if GetToken then exit;
         if Token <> ',' then
           begin
             AddError('Syntax error after arg ' + fd._args_.Names[i]);
             Exit;
           end;
       end;
      if level1(fd._args_.values[i]^) then exit;
    end;
   fData := fd;
   SaveState(@old_state);
   LineIndex := fData.Line;
   LPos := 1;

//   if fData._args_.count > 0 then
//     fData._data_saved_ := fData._args_.Values[0]^
//   else fData._data_saved_.SetValue('');

   fData.Skeep := false;
   fData.fSkeep := false;

//   debug('line: ' + lines.items[lineindex]);

   BlockStack.Push(false, 0, nil);
   Start(fname, nil);
   Result := _data_return_;
   LoadState(@old_state);

   if GetToken then exit;
   if Token <> ')' then
     begin
       AddError('Syntax error for call func: )' + token);
       Exit;
     end;
end;

function TParser.CallIntFunc;
label TO_ERROR;
var i:integer;
    args:TArgs;
begin
   if fData.fSkeep or fData.Skeep then
    begin
      ReadLine;
      LPos := 1;
      exit;
    end;

   args := TArgs.Create;
   Result.SetValue('');

   if GetToken then goto TO_ERROR;
   if Token <> '(' then
     begin
       AddError(Format(ER_FUNC_ARGS_OPEN,[func_map[ind].name]));
       goto TO_ERROR;
     end;

   i := 0;
   if GetToken then goto TO_ERROR;
   while ((i < func_map[ind].count)or(func_map[ind].count = -1))and(Token <> ')') do
    begin
      if i = 0 then
        PutToken;

      if level1(PScData(args.AddArg(''))^) then
       begin
        AddError(Format(ER_FUNC_ARGS_INVALID,[func_map[ind].name]));
        goto TO_ERROR;
       end;

      if GetToken then goto TO_ERROR;
      if (Token <> ',')and(Token <> ')') then
        begin
          AddError(Format(ER_FUNC_ARGS_SYNTAX,[func_map[ind].name]));
          goto TO_ERROR;
        end;
      //if GetToken then goto TO_ERROR;
      inc(i);
    end;
   if(func_map[ind].count <> -1)and(i < func_map[ind].count)then
     begin
       AddError(Format(ER_FUNC_ARGS_COUNT,[func_map[ind].name, func_map[ind].count]));
       goto TO_ERROR;
     end;

   Result := func_map[ind].proc(cgt, el, args);

   if Token <> ')' then
     begin
       AddError(Format(ER_FUNC_ARGS_CLOSE,[func_map[ind].name]));
       goto TO_ERROR;
     end;

TO_ERROR:
  args.Destroy;
end;

function TParser.CallObject;
label TO_ERROR;
var i,mt_ind,count:integer;
    args:TArgs;
    mt:string;
begin
   if fData.fSkeep or fData.Skeep then
    begin
      ReadLine;
      LPos := 1;
      exit;
    end;

   if GetToken then goto TO_ERROR;
   if Token <> '.' then
     begin
       AddError(format(ER_OBJ_FIELD_ACCESS, [obj_map[ind].name]));
       Exit;
     end;

   if GetToken then exit;
   mt := Token;

//   for i := 0 to obj_map[ind].fields_count-1 do  // TODO
//    ;
   mt_ind := -1;
   for i := 0 to obj_map[ind].methods_count-1 do
     if mt = obj_map[ind].methods[i].name then
      begin
       mt_ind := i;
       break;
      end;

   if mt_ind = -1 then
     begin
       AddError(format(ER_OBJ_METHOD_BAD, [RToken, obj_map[ind].name]));
       Exit;
     end
   else count := obj_map[ind].methods[mt_ind].count;

   args := TArgs.Create;
   Result.SetValue('');

   if GetToken then goto TO_ERROR;
   if Token <> '(' then
     begin
       AddError(Format(ER_FUNC_ARGS_OPEN,[mt]));
       goto TO_ERROR;
     end;

   i := 0;
   if GetToken then goto TO_ERROR;
   while ((i < count)or(count = -1))and(Token <> ')') do
    begin
      if i = 0 then
        PutToken;

      if level1(PScData(args.AddArg(''))^) then
       begin
        AddError(Format(ER_FUNC_ARGS_INVALID,[mt]));
        goto TO_ERROR;
       end;

      if GetToken then goto TO_ERROR;
      if (Token <> ',')and(Token <> ')') then
        begin
          AddError(Format(ER_FUNC_ARGS_SYNTAX,[mt]));
          goto TO_ERROR;
        end;
      inc(i);
    end;
   if(count <> -1)and(i < count)then
     begin
       AddError(Format(ER_FUNC_ARGS_COUNT,[mt, count]));
       goto TO_ERROR;
     end;

   Result := obj_map[ind].method_proc(cgt, el, obj_map[ind].obj, mt_ind, args);

   if Token <> ')' then
     begin
       AddError(Format(ER_FUNC_ARGS_CLOSE,[mt]));
       goto TO_ERROR;
     end;

TO_ERROR:
  args.Destroy;
end;

procedure TParser.SumVar;
var r:real;
begin
     case _var_.GetType of
      data_int : _var_.SetValue(_var_.toInt + Var2.toInt);
      data_str : _var_.SetValue(_var_.toStr + Var2.toStr);
      data_real:
       begin
        r := _var_.toReal + Var2.toReal;
        _var_.SetValue(r);
       end;
      data_code: _var_.SetValue(_var_.toCode + Var2.toCode, data_code);
      else _var_.SetValue('');
     end
end;

function TParser.Level1;
begin
   if GetToken then
    begin
     Result := true;
     exit;
    end;

   Result := level2(_var_);
   PutToken;
end;

function TParser.Level2;
var
  var2:TScData;
begin
  Result := Level3(_var_);
  while (Token = '&') do
   begin
    if GetToken then exit;
    Result := Level3(var2);
    //\nDebug('Before cat: var1 = ' + _var_.tocode + ' var2 = ' + var2.toCode);
    _var_.AddValue(var2);
    //\nDebug('Cat: ' + _var_.tocode);
   end;
end;

function TParser.Level3;
var
  Var1,var2:TScData;
  s:string;
begin
  Result := Level4(_var_);
  var1 := _var_;
  while (Token = 'or')or(Token = 'and') do
   begin
    s := Token;
    if GetToken then exit;
    Result := Level4(var2);
    if s = 'or' then
       case Var1.GetType of
        data_int: _var_.SetValue((Var1.toInt <> 0) or (Var2.toInt <> 0));
        data_str: _var_.SetValue((Var1.toStr <> '') or (Var2.toStr <> ''));
        data_real: _var_.SetValue((Var1.toReal <> 0) or (Var2.toReal <> 0));
        else _var_.SetValue(false);
       end
    else
       case Var1.GetType of
        data_int: _var_.SetValue((Var1.toInt <> 0) and (Var2.toInt <> 0));
        data_str: _var_.SetValue((Var1.toStr <> '') and (Var2.toStr <> ''));
        data_real: _var_.SetValue((Var1.toReal <> 0) and (Var2.toReal <> 0));
        else _var_.SetValue(false);
       end
   end;
end;


function TParser.Level4;
var
  Var1,var2:TScData;
  s:string;
begin
  Result := Level5(_var_);
  var1 := _var_;
  while (Token = '=')or(Token = '>')or(Token = '<')or(Token = '>=')or(Token = '<=')or(Token = '!=') do
   begin
    s := Token;
    if GetToken then exit;
    Result := Level5(var2);
    if s = '=' then
       case Var1.GetType of
        data_int: _var_.SetValue(Var1.toInt = Var2.toInt);
        data_str: _var_.SetValue(Var1.toStr = Var2.toStr);
        data_real: _var_.SetValue(Var1.toReal = Var2.toReal);
        else _var_.SetValue(false);
       end
    else if s = '!=' then
       case Var1.GetType of
        data_int: _var_.SetValue(Var1.toInt <> Var2.toInt);
        data_str: _var_.SetValue(Var1.toStr <> Var2.toStr);
        data_real: _var_.SetValue(Var1.toReal <> Var2.toReal);
        else _var_.SetValue(false);
       end
    else if s = '>' then
       case Var1.GetType of
        data_int: _var_.SetValue(Var1.toInt > Var2.toInt);
        //data_str: _var_.SetValue(Var1.toStr = Var2.toStr);
        data_real: _var_.SetValue(Var1.toReal > Var2.toReal);
        else _var_.SetValue(false);
       end
    else if s = '<' then
       case Var1.GetType of
        data_int: _var_.SetValue(Var1.toInt < Var2.toInt);
        //data_str: _var_.SetValue(Var1.toStr = Var2.toStr);
        data_real: _var_.SetValue(Var1.toReal > Var2.toReal);
        else _var_.SetValue(false);
       end
    else if s = '>=' then
       case Var1.GetType of
        data_int: _var_.SetValue(Var1.toInt >= Var2.toInt);
        //data_str: _var_.SetValue(Var1.toStr = Var2.toStr);
        data_real: _var_.SetValue(Var1.toReal >= Var2.toReal);
        else _var_.SetValue(false);
       end
    else if s = '<=' then
       case Var1.GetType of
        data_int: _var_.SetValue(Var1.toInt <= Var2.toInt);
        //data_str: _var_.SetValue(Var1.toStr = Var2.toStr);
        data_real: _var_.SetValue(Var1.toReal >= Var2.toReal);
        else _var_.SetValue(false);
       end;

   end;
end;

function TParser.Level5;
var
  op:byte;
  r:real;
  var2:TScData;
begin
  Result := Level6(_var_);
  while (Token = '-')or(Token = '+') do
   begin
    if Token = '+' then op := 0 else op := 1;
    if GetToken then exit;
    Result := Level6(var2);
    if op = 0 then
     begin
      SumVar(_var_,var2)
     end
    else
     case _var_.GetType of
      data_int : _var_.SetValue(_var_.toInt - Var2.toInt);
      data_str : _var_.SetValue(_var_.toStr);
      data_real:
       begin
        r := _var_.toReal - Var2.toReal;
        _var_.SetValue(r);
       end;
      else _var_.SetValue('');
     end
   end;
end;

function TParser.Level6;
var
    int_var,arg:PScData;
    ind:integer;
    user_name:boolean;
    dt:TScData;
begin
   if TokType = TokCode then
      _var_.SetValue(Token,data_code)
   else if TokType = TokNumber then
      _var_.SetValue(str2int(Token))
   else if TokType = TokSymbol then
    begin
     if Token <> '(' then exit;
      if level1(dt) then exit;
      if GetToken then exit;
      if Token <> ')' then exit;
      _var_ := dt;
    end
   else if TokType = TokName then
      begin
        user_name := false;
        if Token[1] = '_' then
         begin
            if Token = '_data_' then
              if fData._args_.count > 0 then
                _var_ := fData._args_.values[0]^
              else _var_.SetValue('')
            else if Token = '_id_' then
              _var_.SetValue(cgt.elGetCodeName(el))
            else if Token = '_event_count_' then
              _var_.SetValue(_points_count_(pt_event))
            else if Token = '_data_count_' then
              _var_.SetValue(_points_count_(pt_data))
            else if Token = '_var_count_' then
              _var_.SetValue(_points_count_(pt_var))
            else user_name := true;
         end
        else if Token = 'linked' then
          _var_.SetValue(LinkedLexem)
        else if Token = 'str' then
          _var_ := StrLexem
        else if Token = 'code' then
          _var_ := CodeLexem
        else if Token[1] = 'i' then
         begin
            if Token = 'isdef' then
              _var_.SetValue(IsDefLexem)
            else if Token = 'isset' then
              _var_.SetValue(IsSetLexem)
            else if Token = 'isndef' then
              _var_.SetValue(not IsDefLexem)
            else if Token = 'isnset' then
              _var_.SetValue(not IsSetLexem)
            else if Token = 'issec' then
              _var_.SetValue(IsSecLexem)
            else user_name := true;
         end
        else if Token = 'point' then
          ReadCustomValue(_var_)
        else user_name := true;

        if user_name then
          if getInternalVar(token,int_var) then
            _var_ := int_var^
          else if getGlobalVar(token,int_var) then
            _var_ := int_var^
          else if isInternalFunction(token, ind) then
            _var_ := CallIntFunc(ind)
          else if isFunction(token, ind) then
            _var_ := CallFunc(ind)
          else if isObject(token, ind) then
            _var_ := CallObject(ind)
          else
           begin
             arg := fData._args_.find(token);
             if arg <> nil then
               _var_ := arg^
             else ReadValue(RToken,_var_);
           end;
      end
   else _var_.SetValue(Token);

   Result := GetToken;
end;

procedure TParser.ReadValue;
var
    pt,lp:id_point;
    prop:integer;
begin
    pt := cgt.elGetPtName(el, PChar(Name));
    if pt > 0 then
      begin
        if cgt.ptGetType(pt) = pt_event then
          begin
           //\nDebug('Event name found ' + RToken);
           _var_.SetValue(token, data_code);
           exit;
          end;
        lp := cgt.ptGetRLinkPoint(pt);
      end
    else lp := 0;

    if (lp = 0){and(pt = 0)} then
       prop := findProperty(cgt, el, PChar(RToken))
    else prop := -1;

    if(pt = 0)and(lp = 0)and(prop = -1)then
      //case nsection of
       //sec_php : _var_.SetValue('$' + Token + cgt.elGetCodeName(el),data_code);
       {sec_cpp:}  _var_.SetValue(Token + cgt.elGetCodeName(el),data_code)
       //else _var_.SetValue('');
      //end
    else if lp > 0 then
      _var_ := ReadPoint(lp)
    else if (prop <> -1)and((pt = 0) or fData._data_saved_.isEmpty or not cgt.elIsDefProp(el, prop) or (cgt.propGetType(cgt.elGetProperty(el, prop)) = data_comboEx)) then
      begin
       _var_ := readProperty(cgt, cgt.elGetProperty(el, prop));
       //\nDebug('Try to read ' + RToken);
      end
    else
     begin
      //\nDebug('def data: ' + fData._data_.toCode);
      _var_ := fData._data_saved_;
      if null_data then
        fData._data_saved_.SetValue('');
     end;
end;

function TParser.getInternalVar;
begin
   int_var := VarList.find(name);
   Result := int_var <> nil;
end;

function TParser.getGlobalVar;
begin
   gbl_var := GVarList.find(name);
   Result := gbl_var <> nil;
end;

function TParser.isFunction;
var i:integer;
begin
   for i := 0 to funcList.Count-1 do
    if StrIComp(pchar(name), PChar(funclist.items[i])) = 0 then
     begin
       ind := i;
       Result := true;
       exit;
     end;
   Result := false;
end;

function TParser.isInternalFunction;
var i:integer;
begin
   for i := 0 to func_map_size-1 do
    if StrIComp(pchar(name), PChar(func_map[i].name)) = 0 then
     begin
       ind := i;
       Result := true;
       exit;
     end;
   Result := false;
end;

function TParser.isObject;
var i:integer;
begin
   for i := 0 to obj_map_size-1 do
    if StrIComp(pchar(name), PChar(obj_map[i].name)) = 0 then
     begin
       ind := i;
       Result := true;
       exit;
     end;
   Result := false;
end;

procedure TParser.Start;
var var_:PScData;
    ind:integer;
begin
   _err_ := false;
   if ReadLine then exit;

   while not _err_ and not GetToken do
    case TokType of
      TokSymbol:
         AddError(ER_SYNTAX);
      TokName:
          if Token = 'func' then
             FuncLexem(fname,args)
          else if Token = 'end' then
             begin
               if EndLexem then exit;
             end
          else if Token = 'include' then
             IncludeLexem
          else if Token = 'var' then
             VarLexem
          else if Token = 'gvar' then
             GVarLexem
          else if Token = 'return' then
             ReturnLexem
          else if Token = 'for' then
             ForLexem
          else if Token = 'while' then
             WhileLexem
          else if Token = 'if' then
             IfLexem
          else if Token = 'elseif' then
             ElseIfLexem
          else if Token = 'else' then
             ElseLexem
          else if Token = 'print' then
             PrintLexem
          else if Token = 'println' then
             PrintLnLexem
          else if Token = 'event' then
             EventLexem
          else if Token = 'inclevel' then
             IncLevelLexem
          else if Token = 'declevel' then
             DecLevelLexem
          else if Token = 'section' then
             SectionLexem
          //else if Token = 'select' then
          //   SelectLexem
          //else if Token = 'case' then
          //   CaseLexem
          else if getInternalVar(token, var_) then
             VarOpLexem(var_)
          else if getGlobalVar(token, var_) then
             VarOpLexem(var_)
          else if isInternalFunction(token, ind) then
             CallIntFunc(ind)
          else if isObject(token, ind) then
             CallObject(ind)
          else
           if fData.fSkeep then
             begin
               ReadLine;
               LPos := 1;
             end
           else AddError('Undeclared identifier: ' + RToken);
    end;
   //if BlockStack.Count > 0 then
   //  AddError('Key word END IF or NEXT not found');
   //if Funcs.InFunc then
   //  AddError('Key word END not found');
end;

procedure TParser.CallEvent;
var Parser:TParser;
    lp:id_point;
    {i:integer;}
    Args:TArgs;
begin
  lp := cgt.elGetPtName(el, PChar(event));
  if lp > 0 then
   begin
    lp := cgt.ptGetRLinkPoint(lp);
    if lp > 0 then
      begin
        Parser := TParser.Create;
        Args := TArgs.Create;
        Args.AddArg('')^  := data;
        Parser.Parse(cgt, cgt.ptGetParent(lp), lowercase(cgt.ptGetName(lp)), codeb, args, level, nil);
        Args.Destroy;
        Parser.Destroy;
      end;
   end
  else if event <> '' then AddError('Event ' + event + ' not found');
end;

//**************************************************************************************

procedure ConfToCode(const Pack,UName:string);
var
   State:integer;
   List,Body:PStrList;
   i:integer;
   s,MName:string;
begin
   if FileExists(Pack + 'code\hi' + UName + '.hws') then exit;

   List := NewStrList;
   Body := NewStrList;

   State := 0;

   List.LoadFromFile(Pack + 'conf\' + UName + '.ini');
   for i := 0 to List.Count-1 do
    if List.Items[i] <> '' then
     if List.Items[i] = '[Edit]' then
      State := 1
     else if List.Items[i] = '[Property]' then
      State := 2
     else if List.Items[i] = '[Methods]' then
      State := 3
     else if List.Items[i] = '[Type]' then
      State := 4
     else
      case State of
        0:;
        1:;
        2:;
        3:
          begin
             s := List.Items[i];
             MName := getTok(s,'=');
             if MName[1] = '*' then delete(mname,1,1);
             GetTok(s,'|');
             case Str2Int(s) of
               1:
                 begin
                   Body.Add('func ' + MName + '(_data)');
                   Body.Add('  // TODO');
                   Body.Add('end');
                   Body.Add('');
                 end;
               2: ;
               3:
                 begin
                   Body.Add('func ' + MName + '()');
                   Body.Add('  return()');
                   Body.Add('end');
                   Body.Add('');
                 end;
               4: ;
             end;
          end;
        4: ;
      end;
   Body.SaveToFile(Pack + 'code\hi' + UName + '.hws');
   List.free;
   Body.free;
end;

exports
  CreateCode,
  ConfToCode;

end.

